using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace GameDevCalc
{
    public partial class Form1 : Form
    {
        public static System.Globalization.CultureInfo m_InvariantCulture = System.Globalization.CultureInfo.InvariantCulture;

        public enum SELECTED_LIST_BOX { NONE, STACK, MEMORY }
        private SELECTED_LIST_BOX m_SelectedListBox = SELECTED_LIST_BOX.NONE;

        // Dane zapamitane do rozpoczynania Drag&Drop
        Rectangle m_StackMouseRect = Rectangle.Empty;
        int m_StackMouseItemIndex;
        Rectangle m_MemoryMouseRect = Rectangle.Empty;
        int m_MemoryMouseItemIndex;

        VectorFormat m_ActiveVectorFormat = null;

        private enum OP_0
        {
            PI, E,
            // Grupa 3
            RAND,
        }
        private enum OP_1
        {
            FACTORIAL, INT, ROUND, CEIL, FLOOR, FRAC,
            // Grupa 1
            CHANGE_SIGN, ONE_MINUS, RCP, POW2, POW3, SQRT, SATURATE, SIGN,
            // Grupa 2
            LN, LOG2, LOG10, EXP, EXP2, EXP10, SIN, COS, TAN, ASIN, ACOS, ATAN, SINH, COSH, TANH,
            // Grupa 3
            DEGREES, RADIANS, LENGTH, NORMALIZE, RGB, HSB, GRAYSCALE, AVG, ATAN2,
        }
        private enum OP_2
        {
            ADD, SUB, MUL, DIV, MOD, LINEAR,
            // Grupa 1
            POW, MIN, MAX,
            // Grupa 3
            DISTANCE, DOT, CROSS, REFLECT, PROJECT,
        }
        private enum OP_3
        {
            QUADRATIC,
            // Grupa 1
            CLAMP, LERP,
        }

        public Form1()
        {
            InitializeComponent();

            LoadConfiguration();

            UpdateFormatMenu();

            ClearMemory();
        }

        private string GetConfigFileName()
        {
            string Dir = Path.Combine(
                System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData),
                "GameDevCalc");
            if (!Directory.Exists(Dir))
                Directory.CreateDirectory(Dir);
            return Path.Combine(Dir, "GameDevCalc.cfg");

            //return Application.UserAppDataPath + "\\GameDevCalc.cfg";
        }

        private static string CONFIG_HEADER = "GameDevCalc 1";

        private void LoadConfiguration()
        {
            string FileName = GetConfigFileName();
            bool ConfigLoaded = false;
            if (File.Exists(FileName))
            {
                try
                {
                    using (StreamReader Reader = new StreamReader(FileName))
                    {
                        string Header = Reader.ReadLine();
                        if (Header != CONFIG_HEADER)
                            throw new Exception("Invalid file header.");

                        int FormatCount = int.Parse(Reader.ReadLine());
                        for (int i = 0; i < FormatCount; i++)
                        {
                            VectorFormat F = new VectorFormat();
                            F.Format1 = Reader.ReadLine();
                            F.Format2 = Reader.ReadLine();
                            F.Format3 = Reader.ReadLine();
                            F.Format4 = Reader.ReadLine();
                            VectorFormatCollection.Formats.Add(F);
                        }

                        int ActiveFormatIndex = int.Parse(Reader.ReadLine());
                        if (ActiveFormatIndex < 0 || ActiveFormatIndex >= VectorFormatCollection.Formats.Count)
                            m_ActiveVectorFormat = null;
                        else
                            m_ActiveVectorFormat = VectorFormatCollection.Formats[ActiveFormatIndex];
                    }

                    ConfigLoaded = true;
                }
                catch (Exception e)
                {
                    MessageBox.Show(
                        this,
                        "Error while loading configuration file:\n" + FileName + "\n" + e.Message,
                        "Error",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                }
            }

            if (!ConfigLoaded)
            {
                VectorFormatCollection.Formats.Clear();
                VectorFormat Tmp;

                Tmp = new VectorFormat();
                Tmp.Format1 = "%fx";
                Tmp.Format2 = "%fx, %fy";
                Tmp.Format3 = "%fx, %fy, %fz";
                Tmp.Format4 = "%fx, %fy, %fz, %fw";
                VectorFormatCollection.Formats.Add(Tmp);

                Tmp = new VectorFormat();
                Tmp.Format1 = "%fx";
                Tmp.Format2 = "(%fx, %fy)";
                Tmp.Format3 = "(%fx, %fy, %fz)";
                Tmp.Format4 = "(%fx, %fy, %fz, %fw)";
                VectorFormatCollection.Formats.Add(Tmp);

                Tmp = new VectorFormat();
                Tmp.Format1 = "%255ix";
                Tmp.Format2 = "%255ix, %255iy";
                Tmp.Format3 = "%255ix, %255iy, %255iz";
                Tmp.Format4 = "%255ix, %255iy, %255iz, %255iw";
                VectorFormatCollection.Formats.Add(Tmp);

                Tmp = new VectorFormat();
                Tmp.Format1 = "%fxf";
                Tmp.Format2 = "D3DXVECTOR2(%fxf, %fyf)";
                Tmp.Format3 = "D3DXVECTOR3(%fxf, %fyf, %fzf)";
                Tmp.Format4 = "D3DXVECTOR4(%fxf, %fyf, %fzf, %fwf)";
                VectorFormatCollection.Formats.Add(Tmp);

                Tmp = new VectorFormat();
                Tmp.Format1 = "#%255xr";
                Tmp.Format2 = "#%255xr%255xg";
                Tmp.Format3 = "#%255xr%255xg%255xb";
                Tmp.Format4 = "#%255xa%255xr%255xg%255xb";
                VectorFormatCollection.Formats.Add(Tmp);

                Tmp = new VectorFormat();
                Tmp.Format1 = "0x%255xr";
                Tmp.Format2 = "0x%255xr%255xg";
                Tmp.Format3 = "0x%255xr%255xg%255xb";
                Tmp.Format4 = "0x%255xa%255xr%255xg%255xb";
                VectorFormatCollection.Formats.Add(Tmp);

                m_ActiveVectorFormat = VectorFormatCollection.Formats[0];
            }
        }

        void SaveConfiguration()
        {
            string FileName = GetConfigFileName();
            try
            {
                using (StreamWriter Writer = new StreamWriter(FileName))
                {
                    Writer.WriteLine(CONFIG_HEADER);

                    Writer.WriteLine(VectorFormatCollection.Formats.Count);

                    foreach (VectorFormat F in VectorFormatCollection.Formats)
                    {
                        Writer.WriteLine(F.Format1);
                        Writer.WriteLine(F.Format2);
                        Writer.WriteLine(F.Format3);
                        Writer.WriteLine(F.Format4);
                    }

                    if (m_ActiveVectorFormat == null)
                        Writer.WriteLine("-1");
                    else
                        Writer.WriteLine(VectorFormatCollection.Formats.IndexOf(m_ActiveVectorFormat));
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(
                    this,
                    "Error while saving configuration file:\n" + FileName + "\n" + e.Message,
                    "Error",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        private void UpdateFormatMenu()
        {
            while (!(FormatMenu.Items[0] is ToolStripSeparator))
                FormatMenu.Items.RemoveAt(0);

            int Index = 0;
            foreach (VectorFormat F in VectorFormatCollection.Formats)
            {
                ToolStripMenuItem Item = new ToolStripMenuItem(
                    F.GetExampleString(),
                    null,
                    OnMenuFormatClick);
                Item.Tag = F;
                if (m_ActiveVectorFormat == F)
                    Item.Checked = true;
                FormatMenu.Items.Insert(Index, Item);
                Index++;
            }
        }

        private void OnMenuFormatClick(object sender, EventArgs e)
        {
            ToolStripMenuItem Item = sender as ToolStripMenuItem;
            
            m_ActiveVectorFormat = Item.Tag as VectorFormat;

            foreach (ToolStripItem it in FormatMenu.Items)
            {
                if (it is ToolStripMenuItem)
                    (it as ToolStripMenuItem).Checked = (it == Item);
            }
            
            SaveConfiguration();

            UpdateTextBoxFormat();
        }

        private void UpdateTextBoxFormat()
        {
            if (m_ActiveVectorFormat == null)
                TextBoxFormat.Clear();
            else
            {
                Value v = ValueFromComponentTextBoxes();
                if (v == null)
                    TextBoxFormat.Clear();
                else
                    TextBoxFormat.Text = m_ActiveVectorFormat.ValueToString(v);
            }
            TextBoxFormat.Modified = false;
        }

        private void ClearMemory()
        {
            ListMemory.BeginUpdate();
            ListMemory.Items.Clear();
            for (int i = 0; i < 10; i++)
                ListMemory.Items.Add(new Value());
            ListMemory.EndUpdate();
        }

        private void ClearStack()
        {
            ListStack.Items.Clear();
        }

        private void listBox1_DrawItem(object sender, DrawItemEventArgs e)
        {
            if (e.Index == -1) return;

            e.DrawBackground();
            if ((e.State & DrawItemState.Selected) != 0)
                e.DrawFocusRectangle();

            Value v = (Value)(sender as ListBox).Items[e.Index];
            if (v.ComponentCount == 0) return;

            if (v.ComponentCount >= 3)
                using (Brush ColorBr = new SolidBrush(v.Color))
                    e.Graphics.FillRectangle(ColorBr, e.Bounds.Left, e.Bounds.Top, e.Bounds.Height, e.Bounds.Height);

            Brush TextBr = ((e.State & DrawItemState.Selected) != 0 ? Brushes.White : Brushes.Black);
            Rectangle TextRect = new Rectangle(e.Bounds.Left + e.Bounds.Height, e.Bounds.Top, e.Bounds.Width - e.Bounds.Height, e.Bounds.Height);
            e.Graphics.DrawString(v.ToString(), e.Font, TextBr, TextRect, StringFormat.GenericDefault);
        }

        [System.Runtime.InteropServices.DllImport("user32.dll")]
        private static extern int SetScrollPos(int hWnd, int nBar, int nPos, bool bRedraw);
        [System.Runtime.InteropServices.DllImport("user32.dll")]
        private static extern int SendMessage(int hWnd, uint Msg, int wParam, int lParam);

        private void ScrollToBottom(Control C)
        {
            //SetScrollPos((int)ListStack.Handle, 1, int.MaxValue, true);
            // Na podstawie: http://bytes.com/forum/thread262926.html
            SendMessage(
                (int)C.Handle,
                0x0115, // WM_VSCROLL
                (0xFFFF << 16) + (int)ScrollEventType.ThumbPosition,
                0);
        }

        private void NewButton_Click(object sender, EventArgs e)
        {
            Value v = ValueFromAnyTextBoxes();
            if (v != null)
            {
                ListStack.Items.Add(v);
                ScrollToBottom(ListStack);
            }
            SetTextBoxesModifiedToFalse();
        }

        private void SetTextBoxesModifiedToFalse()
        {
            TextBoxX.Modified = false;
            TextBoxY.Modified = false;
            TextBoxZ.Modified = false;
            TextBoxW.Modified = false;
            TextBoxFormat.Modified = false;
            TextBoxComment.Modified = false;
        }

        private Value ValueFromComponentTextBoxes()
        {
            try
            {
                int ComponentCount = 0;
                string S;
                double x = 0.0f, y = 0.0f, z = 0.0f, w = 0.0f;

                S = TextBoxX.Text.Trim();
                if (S.Length == 0)
                    return null;
                x = double.Parse(S, m_InvariantCulture);
                ComponentCount++;

                S = TextBoxY.Text.Trim();
                if (S.Length > 0)
                {
                    y = double.Parse(S, m_InvariantCulture);
                    ComponentCount++;

                    S = TextBoxZ.Text.Trim();
                    if (S.Length > 0)
                    {
                        z = double.Parse(S, m_InvariantCulture);
                        ComponentCount++;

                        S = TextBoxW.Text.Trim();
                        if (S.Length > 0)
                        {
                            w = double.Parse(S, m_InvariantCulture);
                            ComponentCount++;
                        }
                    }
                }

                return new Value(ComponentCount, x, y, z, w);
            }
            catch (Exception)
            {
                return null;
            }
        }

        private void ListStack_SelectedIndexChanged(object sender, EventArgs e)
        {
        }

        private void SelectedValue(Value v)
        {
            if (v.ComponentCount >= 1) TextBoxX.Text = v.x.ToString(m_InvariantCulture);
            else TextBoxX.Clear();
            if (v.ComponentCount >= 2) TextBoxY.Text = v.y.ToString(m_InvariantCulture);
            else TextBoxY.Clear();
            if (v.ComponentCount >= 3) TextBoxZ.Text = v.z.ToString(m_InvariantCulture);
            else TextBoxZ.Clear();
            if (v.ComponentCount >= 4) TextBoxW.Text = v.w.ToString(m_InvariantCulture);
            else TextBoxW.Clear();

            TextBoxX.Modified = false;
            TextBoxY.Modified = false;
            TextBoxZ.Modified = false;
            TextBoxW.Modified = false;

            if (m_ActiveVectorFormat == null)
                TextBoxFormat.Clear();
            else
                TextBoxFormat.Text = m_ActiveVectorFormat.ValueToString(v);
            TextBoxFormat.Modified = false;

            TextBoxComment.Text = v.Comment;
            TextBoxComment.Modified = false;
        }

        private void List_SelectedValueChanged(object sender, EventArgs e)
        {
            ListBox LB = sender as ListBox;

            if (LB == ListStack)
                m_SelectedListBox = SELECTED_LIST_BOX.STACK;
            else if (sender == ListMemory)
                m_SelectedListBox = SELECTED_LIST_BOX.MEMORY;

            if (LB.SelectedIndex != -1)
            {
                Object obj = LB.Items[LB.SelectedIndex];
                if (obj == null) return;
                SelectedValue(obj as Value);
            }
        }

        private void SaveButton_Click(object sender, EventArgs e)
        {
            Value v = ValueFromAnyTextBoxes();

            if (v != null)
            {
                if (m_SelectedListBox == SELECTED_LIST_BOX.STACK)
                {
                    if (ListStack.SelectedIndex != -1)
                    {
                        Value OldV = ListStack.Items[ListStack.SelectedIndex] as Value;
                        if (TextBoxComment.Modified == false)
                            v.Comment = OldV.Comment;
                        ListStack.Items[ListStack.SelectedIndex] = v;
                    }
                }
                else if (m_SelectedListBox == SELECTED_LIST_BOX.MEMORY)
                {
                    Value OldV = ListMemory.Items[ListMemory.SelectedIndex] as Value;
                    if (TextBoxComment.Modified == false)
                        v.Comment = OldV.Comment;
                    if (ListMemory.SelectedIndex != -1)
                        ListMemory.Items[ListMemory.SelectedIndex] = v;
                }
            }

            SetTextBoxesModifiedToFalse();
        }

        // W przypadku bdu pokazuje bd i zwraca null.
        // Jeli zwraca Value, wypenia te jego Comment.
        // Daje focus temu editowi z ktrego pochodz dane.
        private Value ValueFromAnyTextBoxes()
        {
            bool FormatModified = TextBoxFormat.Modified;
            bool ComponentsModified = TextBoxX.Modified || TextBoxY.Modified || TextBoxZ.Modified || TextBoxW.Modified;

            Value v = null;
            bool FromFormatTextBox = false;

            if (FormatModified && ComponentsModified)
            {
                v = ValueFromTextBoxFormat();
                if (v == null)
                {
                    v = ValueFromComponentTextBoxes();
                    if (v == null)
                    {
                        ShowInfoError("Cannot convert - invalid values entered.");
                        return null;
                    }
                }
                else
                    FromFormatTextBox = true;
            }
            else if (FormatModified)
            {
                v = ValueFromTextBoxFormat();
                if (v == null)
                {
                    ShowInfoError("Cannot convert - invalid formatted value entered.");
                    return null;
                }
                FromFormatTextBox = true;
            }
            else
            {
                v = ValueFromComponentTextBoxes();
                if (v == null)
                {
                    ShowInfoError("Cannot convert - invalid component values entered.");
                    return null;
                }
            }

            if (TextBoxComment.Modified)
                v.Comment = TextBoxComment.Text;
            else
            {
                if (FromFormatTextBox)
                    v.Comment = "Entered: " + TextBoxFormat.Text;
                else
                    v.Comment = "Entered: " + v.GetShortString();
            }

            if (FromFormatTextBox)
                TextBoxFormat.Focus();
            else
                TextBoxX.Focus();

            return v;
        }

        // W przypadku niepowodzenia zwraca null.
        private Value ValueFromTextBoxFormat()
        {
            if (m_ActiveVectorFormat == null)
                return null;
            else
                return m_ActiveVectorFormat.StringToValue(TextBoxFormat.Text.Trim());
        }

        private void ButtonClearAll_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show(
                this,
                "Are you sure you want to clear all data?",
                "Clear All",
                MessageBoxButtons.YesNo,
                MessageBoxIcon.Exclamation) == DialogResult.Yes)
            {
                ClearStack();
                ClearMemory();

                TextBoxX.Clear();
                TextBoxY.Clear();
                TextBoxZ.Clear();
                TextBoxW.Clear();
                TextBoxFormat.Clear();
                TextBoxComment.Clear();

                TextBoxX.Modified = false;
                TextBoxY.Modified = false;
                TextBoxZ.Modified = false;
                TextBoxW.Modified = false;
                TextBoxFormat.Modified = false;
                TextBoxComment.Modified = false;
            }
        }

        private void ButtonClearStack_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show(
                this,
                "Are you sure you want to clear the stack?",
                "Clear Stack",
                MessageBoxButtons.YesNo,
                MessageBoxIcon.Exclamation) == DialogResult.Yes)
            {
                ClearStack();
            }
        }

        private void ButtonClearMemory_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show(
                this,
                "Are you sure you want to clear the memory?",
                "Clear Memory",
                MessageBoxButtons.YesNo,
                MessageBoxIcon.Exclamation) == DialogResult.Yes)
            {
                ClearMemory();
            }
        }

        private void SomeButtonMouseEnter(object sender, EventArgs e)
        {
            string s = "";
            if (sender == SaveButton)
                s = "Save (Save modified vector in its place.) [Ctrl+Enter]";
            else if (sender == NewButton)
                s = "New (Add entered vector to the bottom of the stack.) [Enter]";
            else if (sender == TextBoxX)
                s = "X or R";
            else if (sender == TextBoxY)
                s = "Y or G";
            else if (sender == TextBoxZ)
                s = "Z or B";
            else if (sender == TextBoxW)
                s = "W or A";
            else if (sender == InfoLabel)
                s = "Help (Here you can find information about an element under mouse cursor.)";
            else if (sender == DeleteButton)
                s = "Trash (Click or drag and drop a vector here to delete it.) [Del]";
            else if (sender == DuplicateButton)
                s = "Duplicate (Click or drag and drop a vector here to create its copy at the bottom of the stack.) [Alt+=]";
            else if (sender == SeparatorButton)
                s = "Separator (Click to add a separator to the bottom of the stack.) [Alt+-]";
            else if (sender == ButtonClearStack)
                s = "Clear stack (Click to delete all vectors from the stack.)";
            else if (sender == ButtonClearMemory)
                s = "Clear memory (Click to delete all vector from the memory.)";
            else if (sender == ButtonClearAll)
                s = "Clear all (Click to delete all vectors from the stack and the memory.)";
            else if (sender == TextBoxComment)
                s = "Comment (Here you can enter a comment for current vector.)";
            else if (sender == SelectColorButton)
                s = "Select color (Show dialog where you can select a color.) [Alt+~]";
            else if (sender == FormatButton)
                s = "Format (Click to select a vector format or define custom formats.)";
            else if (sender == TextBoxFormat)
                s = "Formatted vector (Here you can read or enter the whole vector formatted in a selected way.)";
            else if (sender == MyHelpButton)
                s = "Help (Click to open documentation text file.)";
            // Grupa gwna
            else if (sender == ButtonAdd) s = "Addition of components of two vectors. [NUM+]";
            else if (sender == ButtonSub) s = "Subtraction of components of two vectors. [NUM-]";
            else if (sender == ButtonMul) s = "Multiplication of components of two vectors. [NUM*]";
            else if (sender == ButtonDiv) s = "Division of components of two vectors. [NUM/]";
            else if (sender == ButtonFactorial) s = "Factorial of components.";
            else if (sender == ButtonMod) s = "Remainder after dividing v1 by v2.";
            else if (sender == ButtonInt) s = "Integer of components.";
            else if (sender == ButtonRound) s = "Components rounded to nearest integer.";
            else if (sender == ButtonCeil) s = "Smallest integers greater than or equal to the components of specified vector.";
            else if (sender == ButtonFloor) s = "Largest integers less than or equal to the components of specified vector.";
            else if (sender == ButtonFrac) s = "Fractional parts of components.";
            else if (sender == ButtonLinear) s = "Returns coefficients (A,B) for linear equation y = ax + b passing through given points (x1,y1), (x2,y2).";
            else if (sender == ButtonQuadratic) s = "Returns coefficients (A,B,C) for quadratic equation y = ax^2 + bx + c passing through given points (x1,y1), (x2,y2), (x3,y3).";
            // Grupa 1
            else if (sender == ButtonChangeSign) s = "Invert sign. Positive numbers become negative, negative become positive.";
            else if (sender == ButtonOneMinus) s = "Invert vector. Every x becomes (1.0 - x).";
            else if (sender == ButtonRcp) s = "Invert vector. Every x becomes (1.0 / x).";
            else if (sender == ButtonPow2) s = "Square. Every x becomes (x * x).";
            else if (sender == ButtonPow3) s = "Cube. Every x becomes (x * x * x).";
            else if (sender == ButtonPow) s = "Power. Raises v1 to v2.";
            else if (sender == ButtonPi) s = "Add PI constant.";
            else if (sender == ButtonE) s = "Add E constant.";
            else if (sender == ButtonSqrt) s = "Square root.";
            else if (sender == ButtonMin) s = "Minimum of components v1, v2.";
            else if (sender == ButtonMax) s = "Maximum of components v1, v2.";
            else if (sender == ButtonClamp) s = "Clamp components of v1 to range v2 ... v3.";
            else if (sender == ButtonLerp) s = "Linear interpolation from v1 to v2 according to v3 = 0.0 ... 1.0.";
            else if (sender == ButtonSaturate) s = "Clamp components of a vector to range 0.0 ... 1.0.";
            else if (sender == ButtonSign) s = "Returns -1, 0 or 1 accordint to sign of vector components.";
            // Grupa 2
            else if (sender == ButtonLn) s = "Natural logarithm (base e) of vector components.";
            else if (sender == ButtonLog2) s = "Logarithm of vector components. Base 2.";
            else if (sender == ButtonLog10) s = "Logarithm of vector components. Base 10.";
            else if (sender == ButtonExp) s = "E raised to the specified power.";
            else if (sender == ButtonExp2) s = "2 raised to the specified power.";
            else if (sender == ButtonExp10) s = "10 raised to the specified power.";
            else if (sender == ButtonSin) s = "Sine of specified angles (in radians).";
            else if (sender == ButtonCos) s = "Cosine of specified angles (in radians).";
            else if (sender == ButtonTan) s = "Tangent of specified angles (in radians).";
            else if (sender == ButtonAsin) s = "Returns angles (in radians) whose sine are specified numbers.";
            else if (sender == ButtonAcos) s = "Returns angles (in radians) whose cosine are specified numbers.";
            else if (sender == ButtonAtan) s = "Returns angles (in radians) whose tangent are specified numbers.";
            else if (sender == ButtonSinh) s = "Hyperbolic sine of specified angles (in radians).";
            else if (sender == ButtonCosh) s = "Hyperbolic cosine of specified angles (in radians).";
            else if (sender == ButtonTanh) s = "Hyperbolic tangent of specified angles (in radians).";
            // Grupa 3
            else if (sender == ButtonDegrees) s = "Converts angle in radians to degrees.";
            else if (sender == ButtonRadians) s = "Converts angle in degrees to radians.";
            else if (sender == ButtonAtan2) s = "Angle of vector (x,y) in radians.";
            else if (sender == ButtonLength) s = "Length of a vector.";
            else if (sender == ButtonDistance) s = "Distance between two points.";
            else if (sender == ButtonNormalize) s = "Scale vector to make its length 1.0.";
            else if (sender == ButtonDot) s = "Dot product of two vectors.";
            else if (sender == ButtonCross) s = "Cross product of two 3D vectors.";
            else if (sender == ButtonReflect) s = "Vector v1 reflected by normal v2.";
            else if (sender == ButtonRgb) s = "Converts (H,S,B) to (R,G,B).";
            else if (sender == ButtonHsb) s = "Converts (R,G,B) to (H,S,B).";
            else if (sender == ButtonGrayscale) s = "Grayscale value of given color.";
            else if (sender == ButtonProject) s = "Vector v1 projected onto v2.";
            else if (sender == ButtonRand) s = "Random numbers in range 0.0 ... 1.0.";
            else if (sender == ButtonAvg) s = "Average of vector components.";

            if (s != null)
            {
                ShowInfoHelp(s);
            }
        }

        private void ShowInfoHelp(string s)
        {
            InfoPanel.BackColor = SystemColors.Info;
            InfoLabel.BackColor = SystemColors.Info;
            InfoLabel.ForeColor = SystemColors.InfoText;
            InfoLabel.Text = s;
        }

        private void ShowInfoError(string Msg)
        {
            InfoPanel.BackColor = Color.Red;
            InfoLabel.BackColor = Color.Red;
            InfoLabel.ForeColor = Color.White;

            InfoLabel.Text = Msg;
        }

        private void ListStack_MouseMove(object sender, MouseEventArgs e)
        {
            OnListMouseMove(sender as ListBox, e.X, e.Y);

            if ((e.Button & MouseButtons.Left) != 0)
            {
                if (m_StackMouseRect != Rectangle.Empty && !m_StackMouseRect.Contains(e.X, e.Y))
                {
                    if (m_StackMouseItemIndex != ListBox.NoMatches &&
                        m_StackMouseItemIndex < ListStack.Items.Count &&
                        ListStack.Items[m_StackMouseItemIndex] is Value)
                    {
                        DoDragDrop(
                            new MyDragData(SELECTED_LIST_BOX.STACK, (Value)ListStack.Items[m_StackMouseItemIndex]),
                            DragDropEffects.Move);
                    }
                }
            }
        }

        private void OnListMouseMove(ListBox LB, int x, int y)
        {
            int Index = LB.IndexFromPoint(x, y);
            if (Index != ListBox.NoMatches && Index < LB.Items.Count)
            {
                if (LB.Items[Index] != null && LB.Items[Index] is Value)
                {
                    string s = (LB.Items[Index] as Value).Comment;
                    if (LB == ListStack)
                        ShowInfoHelp(s);
                    else
                    {
                        int MemNumber = Index + 1;
                        if (MemNumber == 10) MemNumber = 0;
                        ShowInfoHelp(s + " [Alt+" + MemNumber + "]");
                    }
                }
            }
        }

        private void ListStack_MouseDown(object sender, MouseEventArgs e)
        {
            m_StackMouseItemIndex = ListStack.IndexFromPoint(e.X, e.Y);
            if (m_StackMouseItemIndex != ListBox.NoMatches)
            {
                Size DragSize = SystemInformation.DragSize;
                m_StackMouseRect = new Rectangle(
                    new Point(e.X - DragSize.Width / 2, e.Y - DragSize.Height / 2),
                    DragSize);
            }
        }

        private void ListStack_QueryContinueDrag(object sender, QueryContinueDragEventArgs e)
        {
        }

        private void ListStack_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetData(typeof(MyDragData)) != null)
                e.Effect = e.AllowedEffect;
        }

        private void ListMemory_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetData(typeof(MyDragData)) != null)
                e.Effect = e.AllowedEffect;
        }

        private void ListMemory_MouseDown(object sender, MouseEventArgs e)
        {
            m_MemoryMouseItemIndex = ListMemory.IndexFromPoint(e.X, e.Y);
            if (m_MemoryMouseItemIndex != ListBox.NoMatches)
            {
                Size DragSize = SystemInformation.DragSize;
                m_MemoryMouseRect = new Rectangle(
                    new Point(e.X - DragSize.Width / 2, e.Y - DragSize.Height / 2),
                    DragSize);
            }
        }

        private void DeleteButton_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetData(typeof(MyDragData)) != null)
                e.Effect = e.AllowedEffect;
        }

        private void DeleteButton_DockChanged(object sender, EventArgs e)
        {

        }

        private void DeleteButton_DragDrop(object sender, DragEventArgs e)
        {
            object Obj = e.Data.GetData(typeof(MyDragData));
            if (Obj == null) return;
            MyDragData v = Obj as MyDragData;
            if (v.Sender == SELECTED_LIST_BOX.STACK)
            {
                for (int i = 0; i < ListStack.Items.Count; i++)
                {
                    if (ListStack.Items[i] == v.Value)
                    {
                        ListStack.Items.RemoveAt(i);
                        break;
                    }
                }
            }
            else if (v.Sender == SELECTED_LIST_BOX.MEMORY)
            {
                for (int i = 0; i < ListMemory.Items.Count; i++)
                {
                    if (ListMemory.Items[i] == v.Value)
                    {
                        ListMemory.Items[i] = new Value();
                        break;
                    }
                }
            }
        }

        private void DuplicateButton_DragDrop(object sender, DragEventArgs e)
        {
            object Obj = e.Data.GetData(typeof(MyDragData));
            if (Obj == null) return;
            MyDragData v = Obj as MyDragData;
            ListStack.Items.Add(v.Value.Clone());
        }

        private void DuplicateButton_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetData(typeof(MyDragData)) != null)
                e.Effect = e.AllowedEffect;
        }

        private void ListStack_MouseUp(object sender, MouseEventArgs e)
        {
            m_StackMouseRect = Rectangle.Empty;
        }

        private void ListMemory_MouseUp(object sender, MouseEventArgs e)
        {
            m_MemoryMouseRect = Rectangle.Empty;
        }

        private void ListMemory_MouseMove(object sender, MouseEventArgs e)
        {
            OnListMouseMove(sender as ListBox, e.X, e.Y);

            if ((e.Button & MouseButtons.Left) != 0)
            {
                if (m_MemoryMouseRect != Rectangle.Empty && !m_MemoryMouseRect.Contains(e.X, e.Y))
                {
                    if (m_MemoryMouseItemIndex != ListBox.NoMatches &&
                        m_MemoryMouseItemIndex < ListMemory.Items.Count &&
                        ListMemory.Items[m_MemoryMouseItemIndex] is Value)
                    {
                        DoDragDrop(
                            new MyDragData(SELECTED_LIST_BOX.MEMORY, (Value)ListMemory.Items[m_MemoryMouseItemIndex]),
                            DragDropEffects.Move);
                    }
                }
            }
        }

        private void ListMemory_DragDrop(object sender, DragEventArgs e)
        {
            object Obj = e.Data.GetData(typeof(MyDragData));
            if (Obj == null) return;
            MyDragData v = Obj as MyDragData;

            Point ClientPos = ListMemory.PointToClient(new Point(e.X, e.Y));
            int DestIndex = ListMemory.IndexFromPoint(ClientPos);
            if (DestIndex != ListBox.NoMatches && DestIndex < ListMemory.Items.Count)
                ListMemory.Items[DestIndex] = v.Value.Clone();
        }

        private void ListStack_DragDrop(object sender, DragEventArgs e)
        {
            object Obj = e.Data.GetData(typeof(MyDragData));
            if (Obj == null) return;
            MyDragData v = Obj as MyDragData;
            // eby nie przeciga pustych miejsc z pamici
            // (Ale puste miejsca ze stosu mona, bo to separatory, musi si te da je przesuwa.)
            if (v.Sender == SELECTED_LIST_BOX.MEMORY && v.Value.ComponentCount == 0) return;

            int OldIndex = ListStack.Items.IndexOf(v.Value);

            Point ClientPos = ListStack.PointToClient(new Point(e.X, e.Y));
            int DestIndex = ListStack.IndexFromPoint(ClientPos);
            if (DestIndex != ListBox.NoMatches && DestIndex < ListMemory.Items.Count)
                ListStack.Items.Insert(DestIndex > OldIndex ? DestIndex + 1 : DestIndex, v.Value.Clone());
            else
                ListStack.Items.Add(v.Value.Clone());

            ListStack.Items.Remove(v.Value);
        }

        private void SeparatorButton_Click(object sender, EventArgs e)
        {
            ListStack.Items.Add(new Value());
        }

        private void SelectColorButton_Click(object sender, EventArgs e)
        {
            Value v = ValueFromComponentTextBoxes();
            if (v != null)
            {
                Color c = ColorDialog1.Color;
                int R = c.R, G = c.G, B = c.B;
                if (v.ComponentCount >= 1)
                    R = (int)(Math.Round(Math.Max(0.0, Math.Min(255.0, v.x * 255.0))));
                if (v.ComponentCount >= 2)
                    G = (int)(Math.Round(Math.Max(0.0, Math.Min(255.0, v.y * 255.0))));
                if (v.ComponentCount >= 3)
                    B = (int)(Math.Round(Math.Max(0.0, Math.Min(255.0, v.z * 255.0))));
                ColorDialog1.Color = Color.FromArgb(R, G, B);
            }

            if (ColorDialog1.ShowDialog() == DialogResult.OK)
            {
                Color C = ColorDialog1.Color;

                double R = (double)C.R / 255.0;
                double G = (double)C.G / 255.0;
                double B = (double)C.B / 255.0;

                TextBoxX.Text = R.ToString(m_InvariantCulture);
                TextBoxY.Text = G.ToString(m_InvariantCulture);
                TextBoxZ.Text = B.ToString(m_InvariantCulture);
                TextBoxComment.Text = "Selected color: (" + C.R + "," + C.G + "," + C.B + ")";

                TextBoxX.Modified = true;
                TextBoxY.Modified = true;
                TextBoxZ.Modified = true;
                TextBoxComment.Modified = true;
            }
        }

        private void ButtonOpClick(object sender, EventArgs e)
        {
            if (sender == ButtonAdd) Op2(OP_2.ADD);
            else if (sender == ButtonSub) Op2(OP_2.SUB);
            else if (sender == ButtonMul) Op2(OP_2.MUL);
            else if (sender == ButtonDiv) Op2(OP_2.DIV);
            else if (sender == ButtonFactorial) Op1(OP_1.FACTORIAL);
            else if (sender == ButtonMod) Op2(OP_2.MOD);
            else if (sender == ButtonInt) Op1(OP_1.INT);
            else if (sender == ButtonRound) Op1(OP_1.ROUND);
            else if (sender == ButtonCeil) Op1(OP_1.CEIL);
            else if (sender == ButtonFloor) Op1(OP_1.FLOOR);
            else if (sender == ButtonFrac) Op1(OP_1.FRAC);
            else if (sender == ButtonLinear) Op2(OP_2.LINEAR);
            else if (sender == ButtonQuadratic) Op3(OP_3.QUADRATIC);
            // Grupa 1
            else if (sender == ButtonChangeSign) Op1(OP_1.CHANGE_SIGN);
            else if (sender == ButtonOneMinus) Op1(OP_1.ONE_MINUS);
            else if (sender == ButtonRcp) Op1(OP_1.RCP);
            else if (sender == ButtonPow2) Op1(OP_1.POW2);
            else if (sender == ButtonPow3) Op1(OP_1.POW3);
            else if (sender == ButtonPow) Op2(OP_2.POW);
            else if (sender == ButtonPi) Op0(OP_0.PI);
            else if (sender == ButtonE) Op0(OP_0.E);
            else if (sender == ButtonSqrt) Op1(OP_1.SQRT);
            else if (sender == ButtonMin) Op2(OP_2.MIN);
            else if (sender == ButtonMax) Op2(OP_2.MAX);
            else if (sender == ButtonClamp) Op3(OP_3.CLAMP);
            else if (sender == ButtonLerp) Op3(OP_3.LERP);
            else if (sender == ButtonSaturate) Op1(OP_1.SATURATE);
            else if (sender == ButtonSign) Op1(OP_1.SIGN);
            // Grupa 2
            else if (sender == ButtonLn) Op1(OP_1.LN);
            else if (sender == ButtonLog2) Op1(OP_1.LOG2);
            else if (sender == ButtonLog10) Op1(OP_1.LOG10);
            else if (sender == ButtonExp) Op1(OP_1.EXP);
            else if (sender == ButtonExp2) Op1(OP_1.EXP2);
            else if (sender == ButtonExp10) Op1(OP_1.EXP10);
            else if (sender == ButtonSin) Op1(OP_1.SIN);
            else if (sender == ButtonCos) Op1(OP_1.COS);
            else if (sender == ButtonTan) Op1(OP_1.TAN);
            else if (sender == ButtonAsin) Op1(OP_1.ASIN);
            else if (sender == ButtonAcos) Op1(OP_1.ACOS);
            else if (sender == ButtonAtan) Op1(OP_1.ATAN);
            else if (sender == ButtonSinh) Op1(OP_1.SINH);
            else if (sender == ButtonCosh) Op1(OP_1.COSH);
            else if (sender == ButtonTanh) Op1(OP_1.TANH);
            // Grupa 3
            else if (sender == ButtonDegrees) Op1(OP_1.DEGREES);
            else if (sender == ButtonRadians) Op1(OP_1.RADIANS);
            else if (sender == ButtonAtan2) Op1(OP_1.ATAN2);
            else if (sender == ButtonLength) Op1(OP_1.LENGTH);
            else if (sender == ButtonDistance) Op2(OP_2.DISTANCE);
            else if (sender == ButtonNormalize) Op1(OP_1.NORMALIZE);
            else if (sender == ButtonDot) Op2(OP_2.DOT);
            else if (sender == ButtonCross) Op2(OP_2.CROSS);
            else if (sender == ButtonReflect) Op2(OP_2.REFLECT);
            else if (sender == ButtonRgb) Op1(OP_1.RGB);
            else if (sender == ButtonHsb) Op1(OP_1.HSB);
            else if (sender == ButtonGrayscale) Op1(OP_1.GRAYSCALE);
            else if (sender == ButtonProject) Op2(OP_2.PROJECT);
            else if (sender == ButtonRand) Op0(OP_0.RAND);
            else if (sender == ButtonAvg) Op1(OP_1.AVG);
        }

        void Op0(OP_0 Op)
        {
            int ComponentCount = 1;
            double[] Result = new double[4];
            string Comment = "";

            switch (Op)
            {
                case OP_0.PI:
                    ComponentCount = 1;
                    Result[0] = Math.PI;
                    Comment = "PI";
                    break;
                case OP_0.E:
                    ComponentCount = 1;
                    Result[0] = Math.E;
                    Comment = "E";
                    break;
                case OP_0.RAND:
                    ComponentCount = 4;
                    Random Rand = new Random();
                    Result[0] = Rand.NextDouble();
                    Result[1] = Rand.NextDouble();
                    Result[2] = Rand.NextDouble();
                    Result[3] = Rand.NextDouble();
                    Comment = "Random";
                    break;
                default:
                    ShowInfoError("Internal error: Invalid OP_0.");
                    return;
            }

            for (int i = 0; i < ComponentCount; i++)
            {
                if (double.IsInfinity(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
                if (double.IsNaN(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
            }

            Value ResultV = new Value(ComponentCount, Result[0], Result[1], Result[2], Result[3]);
            ResultV.Comment = Comment;
            ListStack.Items.Add(ResultV);
        }

        void Op1(OP_1 Op)
        {
            object Obj;
/*            if (m_SelectedListBox == SELECTED_LIST_BOX.STACK &&
                ListStack.SelectedIndex > -1 &&
                ListStack.SelectedIndex < ListStack.Items.Count)
            {
                Obj = ListStack.Items[ListStack.SelectedIndex];
            }
            else*/
            if (ListStack.Items.Count > 0)
                Obj = ListStack.Items[ListStack.Items.Count - 1];
            else
            {
                ShowInfoError("No vectors on the stack.");
                return;
            }

            if (Obj == null || !(Obj is Value))
            {
                ShowInfoError("Cannot perform operation.");
                return;
            }

            Value v = Obj as Value;
            if (v.ComponentCount == 0)
            {
                ShowInfoError("Cannot perform operation.");
                return;
            }

            int ComponentCount = v.ComponentCount;
            double[] Result = new double[4];
            string Comment = "";

            switch (Op)
            {
                case OP_1.FACTORIAL:
                    for (int i = 0; i < v.ComponentCount; i++)
                    {
                        if (!CalcFactorial(out Result[i], (int)v[i]))
                        {
                            ShowInfoError("Cannot calculate factorial.");
                            return;
                        }
                    }
                    Comment = "factorial(" + v.GetShortString() + ")";
                    break;
                case OP_1.INT:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = (double)(int)v[i];
                    Comment = "int(" + v.GetShortString() + ")";
                    break;
                case OP_1.ROUND:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Round(v[i]);
                    Comment = "round(" + v.GetShortString() + ")";
                    break;
                case OP_1.CEIL:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Ceiling(v[i]);
                    Comment = "ceil(" + v.GetShortString() + ")";
                    break;
                case OP_1.FLOOR:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Floor(v[i]);
                    Comment = "floor(" + v.GetShortString() + ")";
                    break;
                case OP_1.FRAC:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = CalcFrac(v[i]);
                    Comment = "frac(" + v.GetShortString() + ")";
                    break;
                // Grupa 1
                case OP_1.CHANGE_SIGN:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = -v[i];
                    Comment = "- (" + v.GetShortString() + ")";
                    break;
                case OP_1.ONE_MINUS:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = 1.0f - v[i];
                    Comment = "1 - (" + v.GetShortString() + ")";
                    break;
                case OP_1.RCP:
                    for (int i = 0; i < v.ComponentCount; i++)
                    {
                        if (v[i] == 0.0f)
                        {
                            ShowInfoError("Division by zero.");
                            return;
                        }
                        Result[i] = 1.0f / v[i];
                    }
                    Comment = "1 - (" + v.GetShortString() + ")";
                    break;
                case OP_1.POW2:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = v[i] * v[i];
                    Comment = "(" + v.GetShortString() + ") ^ 2";
                    break;
                case OP_1.POW3:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = v[i] * v[i] * v[i];
                    Comment = "(" + v.GetShortString() + ") ^ 3";
                    break;
                case OP_1.SQRT:
                    for (int i = 0; i < v.ComponentCount; i++)
                    {
                        if (v[i] < 0.0f)
                        {
                            ShowInfoError("Square root of negative number.");
                            return;
                        }
                        Result[i] = Math.Sqrt(v[i]);
                    }
                    Comment = "sqrt(" + v.GetShortString() + ")";
                    break;
                case OP_1.SATURATE:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Min(1.0f, Math.Max(0.0f, v[i]));
                    Comment = "saturate(" + v.GetShortString() + ")";
                    break;
                case OP_1.SIGN:
                    for (int i = 0; i < v.ComponentCount; i++)
                    {
                        if (v[i] > 0.0f) Result[i] = 1.0f;
                        else if (v[i] == 0.0f) Result[i] = 0.0f;
                        else Result[i] = -1.0f;
                    }
                    Comment = "sign(" + v.GetShortString() + ")";
                    break;
                // Grupa 2
                case OP_1.LN:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Log(v[i]);
                    Comment = "ln(" + v.GetShortString() + ")";
                    break;
                case OP_1.LOG2:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Log(v[i], 2);
                    Comment = "log2(" + v.GetShortString() + ")";
                    break;
                case OP_1.LOG10:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Log10(v[i]);
                    Comment = "log10(" + v.GetShortString() + ")";
                    break;
                case OP_1.EXP:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Exp(v[i]);
                    Comment = "exp(" + v.GetShortString() + ")";
                    break;
                case OP_1.EXP2:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Pow(2.0f, v[i]);
                    Comment = "log2(" + v.GetShortString() + ")";
                    break;
                case OP_1.EXP10:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Pow(10.0f, v[i]);
                    Comment = "exp10(" + v.GetShortString() + ")";
                    break;
                case OP_1.SIN:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Sin(v[i]);
                    Comment = "sin(" + v.GetShortString() + ")";
                    break;
                case OP_1.COS:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Cos(v[i]);
                    Comment = "cos(" + v.GetShortString() + ")";
                    break;
                case OP_1.TAN:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Tan(v[i]);
                    Comment = "tan(" + v.GetShortString() + ")";
                    break;
                case OP_1.ASIN:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Asin(v[i]);
                    Comment = "asin(" + v.GetShortString() + ")";
                    break;
                case OP_1.ACOS:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Acos(v[i]);
                    Comment = "acos(" + v.GetShortString() + ")";
                    break;
                case OP_1.ATAN:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Atan(v[i]);
                    Comment = "atan(" + v.GetShortString() + ")";
                    break;
                case OP_1.SINH:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Sinh(v[i]);
                    Comment = "sinh(" + v.GetShortString() + ")";
                    break;
                case OP_1.COSH:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Cosh(v[i]);
                    Comment = "cosh(" + v.GetShortString() + ")";
                    break;
                case OP_1.TANH:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = Math.Tanh(v[i]);
                    Comment = "tanh(" + v.GetShortString() + ")";
                    break;
                // Grupa 3
                case OP_1.DEGREES:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = (v[i] / Math.PI * 180.0);
                    Comment = "degrees(" + v.GetShortString() + ")";
                    break;
                case OP_1.RADIANS:
                    for (int i = 0; i < v.ComponentCount; i++)
                        Result[i] = (v[i] / 180.0 * Math.PI);
                    Comment = "radians(" + v.GetShortString() + ")";
                    break;
                case OP_1.LENGTH:
                    double R = 0.0f;
                    for (int i = 0; i < v.ComponentCount; i++)
                        R += v[i] * v[i];
                    ComponentCount = 1;
                    Result[0] = Math.Sqrt(R);
                    Comment = "length(" + v.GetShortString() + ")";
                    break;
                case OP_1.NORMALIZE:
                    double Length = 0.0f;
                    for (int i = 0; i < v.ComponentCount; i++)
                        Length += v[i] * v[i];
                    Length = Math.Sqrt(Length);
                    if (Length == 0.0f)
                    {
                        ShowInfoError("Vector has zero length.");
                        return;
                    }
                    for (int i = 0; i < ComponentCount; i++)
                        Result[i] = v[i] / Length;
                    Comment = "normalize(" + v.GetShortString() + ")";
                    break;
                case OP_1.RGB:
                    if (v.ComponentCount < 3)
                    {
                        ShowInfoError("Requires input (H,S,B).");
                        return;
                    }
                    CalcHsbToRgb(
                        out Result[0], out Result[1], out Result[2],
                        v.x, v.y, v.z);
                    if (v.ComponentCount == 4)
                    {
                        ComponentCount = 4;
                        Result[3] = v[3];
                    }
                    else
                        ComponentCount = 3;
                    Comment = "rgb(" + v.GetShortString() + ")";
                    break;
                case OP_1.HSB:
                    if (v.ComponentCount < 3)
                    {
                        ShowInfoError("Requires input (R,G,B).");
                        return;
                    }
                    CalcRgbToHsb(
                        out Result[0], out Result[1], out Result[2],
                        v.x, v.y, v.z);
                    if (v.ComponentCount == 4)
                    {
                        ComponentCount = 4;
                        Result[3] = v[3];
                    }
                    else
                        ComponentCount = 3;
                    Comment = "hsb(" + v.GetShortString() + ")";
                    break;
                case OP_1.GRAYSCALE:
                    if (v.ComponentCount < 3)
                    {
                        ShowInfoError("Requires input (R,G,B).");
                        return;
                    }
                    ComponentCount = 1;
                    Result[0] = 0.299 * v[0] + 0.587 * v[1] + 0.114 * v[2];
                    Comment = "grayscale(" + v.GetShortString() + ")";
                    break;
                case OP_1.AVG:
                    ComponentCount = 1;
                    Result[0] = v[0];
                    if (v.ComponentCount > 1)
                    {
                        for (int i = 1; i < v.ComponentCount; i++)
                            Result[0] += v[i];
                        Result[0] /= (double)v.ComponentCount;
                    }
                    Comment = "avg(" + v.GetShortString() + ")";
                    break;
                case OP_1.ATAN2:
                    if (v.ComponentCount < 2)
                    {
                        ShowInfoError("Requires input (X,Y).");
                        return;
                    }
                    ComponentCount = 1;
                    Result[0] = Math.Atan2(v[1], v[0]);
                    Comment = "atan2(" + v.GetShortString() + ")";
                    break;
                default:
                    ShowInfoError("Internal error: Invalid OP_1.");
                    return;
            }

            for (int i = 0; i < ComponentCount; i++)
            {
                if (double.IsInfinity(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
                if (double.IsNaN(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
            }

            Value ResultV = new Value(ComponentCount, Result[0], Result[1], Result[2], Result[3]);
            ResultV.Comment = Comment;
            ListStack.Items.Add(ResultV);
        }

        private void CalcRgbToHsb(out double Hue, out double Saturation, out double Brightness, double R, double G, double B)
        {
            double Min = Math.Min(R, Math.Min(G, B));
            double Max = Math.Max(R, Math.Max(G, B));

            Brightness = Max;

            if (Max == 0.0)
                Saturation = 0.0;
            else
                Saturation = 1.0 - (Min / Max);

            if (Max == Min)
                Hue = 0.0; // Undefined
            else if (Max == R && G >= B)
                Hue = (60.0 / 360.0) * (G - B) / (Max - Min);
            else if (Max == R) // && G < B
                Hue = (60.0 / 360.0) * (G - B) / (Max - Min) + 1.0;
            else if (Max == G)
                Hue = (60.0 / 360.0) * (B - R) / (Max - Min) + (120.0 / 360.0);
            else // Max == B
                Hue = (60.0 / 360.0) * (R - G) / (Max - Min) + (240.0 / 360.0);
        }

        private void CalcHsbToRgb(out double R, out double G, out double B, double Hue, double Saturation, double Brightness)
        {
            double h = CalcFrac(Hue) * 6.0;
            double f = CalcFrac(h);
            double p = Brightness * (1.0 - Saturation);
            double q = Brightness * (1.0 - Saturation * f);
            double t = Brightness * (1.0 - (Saturation * (1.0 - f)));

            switch ((int)h)
            {
                case 0:
                    R = Brightness;
                    G = t;
                    B = p;
                    return;
                case 1:
                    R = q;
                    G = Brightness;
                    B = p;
                    return;
                case 2:
                    R = p;
                    G = Brightness;
                    B = t;
                    return;
                case 3:
                    R = p;
                    G = q;
                    B = Brightness;
                    return;
                case 4:
                    R = t;
                    G = p;
                    B = Brightness;
                    return;
                default:
                    R = Brightness;
                    G = p;
                    B = q;
                    return;
            }
        }

        private double CalcFrac(double x)
        {
            return x - CalcTrunc(x);
        }

        private double CalcTrunc(double x)
        {
            return (x < 0.0) ? Math.Ceiling(x) : Math.Floor(x);
        }

        private bool CalcFactorial(out double Result, int X)
        {
            Result = 0.0f; // Compiler's Food
            if (X < 0) return false;
            else if (X < 2) Result = 1.0f;
            else if (X > 170) return false;
            else
            {
                Result = 1.0f;
                double fi = 2.0f;
                for (int i = 2; i <= X; i++)
                {
                    Result *= fi;
                    fi += 1.0f;
                }
            }

            return true;
        }

        private void Op2(OP_2 Op)
        {
            object Obj1, Obj2;
            if (ListStack.Items.Count > 1)
            {
                Obj1 = ListStack.Items[ListStack.Items.Count - 2];
                Obj2 = ListStack.Items[ListStack.Items.Count - 1];
            }
            else
                return;

            if (Obj1 == null || !(Obj1 is Value) ||
                Obj2 == null || !(Obj2 is Value))
            {
                ShowInfoError("Cannot perform operation.");
                return;
            }

            Value v1 = Obj1 as Value;
            Value v2 = Obj2 as Value;
            if (v1.ComponentCount == 0 || v2.ComponentCount == 0)
            {
                ShowInfoError("Cannot perform operation.");
                return;
            }

            int ComponentCount = Math.Min(v1.ComponentCount, v2.ComponentCount);
            double[] Result = new double[4];
            string Comment = "";

            switch (Op)
            {
                case OP_2.ADD:
                    for (int i = 0; i < ComponentCount; i++)
                        Result[i] = v1[i] + v2[i];
                    Comment = "(" + v1.GetShortString() + ") + (" + v2.GetShortString() + ")";
                    break;
                case OP_2.SUB:
                    for (int i = 0; i < ComponentCount; i++)
                        Result[i] = v1[i] - v2[i];
                    Comment = "(" + v1.GetShortString() + ") - (" + v2.GetShortString() + ")";
                    break;
                case OP_2.MUL:
                    if (v1.ComponentCount == 1)
                    {
                        ComponentCount = v2.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = v1[0] * v2[i];
                    }
                    else if (v2.ComponentCount == 1)
                    {
                        ComponentCount = v1.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = v1[i] * v2[0];
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = v1[i] * v2[i];
                    }
                    Comment = "(" + v1.GetShortString() + ") * (" + v2.GetShortString() + ")";
                    break;
                case OP_2.DIV:
                    if (v2.ComponentCount == 1)
                    {
                        ComponentCount = v1.ComponentCount;
                        if (v2[0] == 0.0f)
                        {
                            ShowInfoError("Division by zero.");
                            return;
                        }
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = v1[i] / v2[0];
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                        {
                            if (v2[i] == 0.0f)
                            {
                                ShowInfoError("Division by zero.");
                                return;
                            }
                            Result[i] = v1[i] / v2[i];
                        }
                    }
                    Comment = "(" + v1.GetShortString() + ") / (" + v2.GetShortString() + ")";
                    break;
                case OP_2.LINEAR:
                    if (v1.ComponentCount != 2 || v2.ComponentCount != 2)
                    {
                        ShowInfoError("Operands must be (x1,y1), (x2,y2).");
                        return;
                    }
                    double Denominator = v2.x - v1.x;
                    if (Denominator == 0.0f)
                    {
                        ShowInfoError("Division by zero - vertical line.");
                        return;
                    }
                    ComponentCount = 2;
                    Result[0] = (v2.y - v1.y) / Denominator;
                    Result[1] = (v2.x * v1.y - v2.y * v1.x) / Denominator;
                    Comment = "linear( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    break;
                case OP_2.POW:
                    if (v2.ComponentCount == 1)
                    {
                        ComponentCount = v1.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Pow(v1[i], v2[0]);
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Pow(v1[i], v2[i]);
                    }
                    Comment = "pow( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    break;
                case OP_2.MIN:
                    if (v2.ComponentCount == 1)
                    {
                        ComponentCount = v1.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Min(v1[i], v2[0]);
                    }
                    else if (v1.ComponentCount == 1)
                    {
                        ComponentCount = v2.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Min(v1[0], v2[i]);
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Min(v1[i], v2[i]);
                    }
                    Comment = "min( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    break;
                case OP_2.MAX:
                    if (v2.ComponentCount == 1)
                    {
                        ComponentCount = v1.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Max(v1[i], v2[0]);
                    }
                    else if (v1.ComponentCount == 1)
                    {
                        ComponentCount = v2.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Max(v1[0], v2[i]);
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Max(v1[i], v2[i]);
                    }
                    Comment = "max( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    break;
                // Grupa 3
                case OP_2.DISTANCE:
                    {
                        int InputCount = ComponentCount;
                        ComponentCount = 1;
                        Result[0] = 0.0f;
                        for (int i = 0; i < InputCount; i++)
                            Result[0] += (v2[i] - v1[i]) * (v2[i] - v1[i]);
                        Result[0] = Math.Sqrt(Result[0]);
                        Comment = "distance( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    }
                    break;
                case OP_2.DOT:
                    {
                        int InputCount = ComponentCount;
                        ComponentCount = 1;
                        Result[0] = 0.0f;
                        for (int i = 0; i < InputCount; i++)
                            Result[0] += v1[i] * v2[i];
                        Comment = "dot( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    }
                    break;
                case OP_2.CROSS:
                    {
                        if (v1.ComponentCount < 3 || v2.ComponentCount < 3)
                        {
                            ShowInfoError("Cross product requires vectors (x1,y1,z1), (x2,y2,z2).");
                            return;
                        }
                        ComponentCount = 3;
                        Result[0] = v1.y * v2.z - v1.z * v2.y;
                        Result[1] = v1.z * v2.x - v1.x * v2.z;
                        Result[2] = v1.x * v2.y - v1.y * v2.x;
                        Comment = "cross( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    }
                    break;
                case OP_2.REFLECT:
                    {
                        if (v1.ComponentCount < 3 || v2.ComponentCount < 3)
                        {
                            ShowInfoError("Reflect requires vectors (x1,y1,z1), (x2,y2,z2).");
                            return;
                        }
                        ComponentCount = 3;
                        double f = 2.0 * CalcDot(v1.x, v1.y, v1.z, v2.x, v2.y, v2.z);
                        Result[0] = v1.x - f * v2.x;
                        Result[1] = v1.y - f * v2.y;
                        Result[2] = v1.z - f * v2.z;
                        Comment = "reflect( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    }
                    break;
                case OP_2.PROJECT:
                    {
                        if (v1.ComponentCount < 3 || v2.ComponentCount < 3)
                        {
                            ShowInfoError("Project requires vectors (x1,y1,z1), (x2,y2,z2).");
                            return;
                        }
                        ComponentCount = 3;
                        double T = (v1.x * v2.x + v1.y * v2.y + v1.z * v2.z) / (v2.x * v2.x + v2.y * v2.y + v2.z * v2.z);
                        Result[0] = v2.x * T;
                        Result[1] = v2.y * T;
                        Result[2] = v2.z * T;
                        Comment = "project( (" + v1.GetShortString() + "), (" + v2.GetShortString() + ") )";
                    }
                    break;
                default:
                    ShowInfoError("Internal error: Invalid OP_2.");
                    return;
            }

            for (int i = 0; i < ComponentCount; i++)
            {
                if (double.IsInfinity(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
                if (double.IsNaN(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
            }

            Value ResultV = new Value(ComponentCount, Result[0], Result[1], Result[2], Result[3]);
            ResultV.Comment = Comment;
            ListStack.Items.Add(ResultV);
        }

        private double CalcDot(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            return x1 * x2 + y1 * y2 + z1 * z2;
        }

        private void Op3(OP_3 Op)
        {
            object Obj1, Obj2, Obj3;
            if (ListStack.Items.Count > 2)
            {
                Obj1 = ListStack.Items[ListStack.Items.Count - 3];
                Obj2 = ListStack.Items[ListStack.Items.Count - 2];
                Obj3 = ListStack.Items[ListStack.Items.Count - 1];
            }
            else
                return;

            if (Obj1 == null || !(Obj1 is Value) ||
                Obj2 == null || !(Obj2 is Value) ||
                Obj3 == null || !(Obj3 is Value))
            {
                ShowInfoError("Cannot perform operation.");
                return;
            }

            Value v1 = Obj1 as Value;
            Value v2 = Obj2 as Value;
            Value v3 = Obj3 as Value;
            if (v1.ComponentCount == 0 || v2.ComponentCount == 0 || v3.ComponentCount == 0)
            {
                ShowInfoError("Cannot perform operation.");
                return;
            }

            int ComponentCount = Math.Min(Math.Min(v1.ComponentCount, v2.ComponentCount), v3.ComponentCount);
            double[] Result = new double[4];
            string Comment = "";

            switch (Op)
            {
                case OP_3.QUADRATIC:
                    if (v1.ComponentCount != 2 || v2.ComponentCount != 2 || v3.ComponentCount != 2)
                    {
                        ShowInfoError("Operands must be (x1,y1), (x2,y2), (x3,y3).");
                        return;
                    }
                    double x1 = v1.x, x2 = v2.x, x3 = v3.x;
                    double y1 = v1.y, y2 = v2.y, y3 = v3.y;
                    double W = x1 * x1 * x2 + x3 * x3 * x1 + x2 * x2 * x3 - x1 * x1 * x3 - x2 * x2 * x1 - x3 * x3 * x2;
                    if (W == 0.0f)
                    {
                        ShowInfoError("Division by zero.");
                        return;
                    }
                    double A = y1 * x2 + y3 * x1 + y2 * x3 - y1 * x3 - y2 * x1 - y3 * x2;
                    double B = x1 * x1 * y2 + x3 * x3 * y1 + x2 * x2 * y3 - x1 * x1 * y3 - x2 * x2 * y1 - x3 * x3 * y2;
                    double C = x1 * x1 * x2 * y3 + x3 * x3 * x1 * y2 + x2 * x2 * x3 * y1 - x1 * x1 * x3 * y2 - x2 * x2 * x1 * y3 - x3 * x3 * x2 * y1;
                    ComponentCount = 3;
                    Result[0] = A / W;
                    Result[1] = B / W;
                    Result[2] = C / W;

                    Comment = "quadratic( (" + v1.GetShortString() + "), (" + v2.GetShortString() + "), (" + v3.GetShortString() + ") )";
                    break;
                case OP_3.CLAMP:
                    if (v2.ComponentCount == 1 && v3.ComponentCount == 1)
                    {
                        ComponentCount = v1.ComponentCount;
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Min(v3[0], Math.Max(v2[0], v1[i]));
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = Math.Min(v3[i], Math.Max(v2[i], v1[i]));
                    }
                    Comment = "clamp( (" + v1.GetShortString() + "), (" + v2.GetShortString() + "), (" + v3.GetShortString() + ") )";
                    break;
                case OP_3.LERP:
                    if (v3.ComponentCount == 1)
                    {
                        ComponentCount = Math.Min(v1.ComponentCount, v2.ComponentCount);
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = v1[i] * (1.0f - v3[0]) + v2[i] * (v3[0]);
                    }
                    else
                    {
                        for (int i = 0; i < ComponentCount; i++)
                            Result[i] = v1[i] * (1.0f - v3[i]) + v2[i] * (v3[i]);
                    }
                    Comment = "lerp( (" + v1.GetShortString() + "), (" + v2.GetShortString() + "), (" + v3.GetShortString() + ") )";
                    break;
                default:
                    ShowInfoError("Internal error: Invalid OP_3.");
                    return;
            }

            for (int i = 0; i < ComponentCount; i++)
            {
                if (double.IsInfinity(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
                if (double.IsNaN(Result[i]))
                {
                    ShowInfoError("Infinity.");
                    return;
                }
            }

            Value ResultV = new Value(ComponentCount, Result[0], Result[1], Result[2], Result[3]);
            ResultV.Comment = Comment;
            ListStack.Items.Add(ResultV);
        }

        private void toolStripMenuItem1_Click(object sender, EventArgs e)
        {
            FormFormats Form = new FormFormats();
            Form.ShowDialog(this);

            if (m_ActiveVectorFormat != null)
            {
                if (VectorFormatCollection.Formats.IndexOf(m_ActiveVectorFormat) == -1)
                {
                    if (VectorFormatCollection.Formats.Count == 0)
                        m_ActiveVectorFormat = null;
                    else
                        m_ActiveVectorFormat = VectorFormatCollection.Formats[0];
                }
            }

            SaveConfiguration();

            UpdateFormatMenu();

            UpdateTextBoxFormat();
        }

        private void FormatButton_Click(object sender, EventArgs e)
        {
            FormatMenu.Show(FormatButton, new Point(0, FormatButton.Height));
        }

        private void MyHelpButton_Click(object sender, EventArgs e)
        {
            OnHelp();
        }

        public static void OnHelp()
        {
            string HelpFileName = Path.Combine(Application.StartupPath, "GameDevCalc.txt");
            try
            {
                System.Diagnostics.Process.Start(HelpFileName);
            }
            catch (Exception)
            {
            }
        }

        private void DeleteButton_Click(object sender, EventArgs e)
        {
            if (m_SelectedListBox == SELECTED_LIST_BOX.STACK)
            {
                int SelectedIndex = ListStack.SelectedIndex;
                if (SelectedIndex >= 0 && SelectedIndex < ListStack.Items.Count)
                    ListStack.Items.RemoveAt(SelectedIndex);
            }
            else if (m_SelectedListBox == SELECTED_LIST_BOX.MEMORY)
            {
                int SelectedIndex = ListMemory.SelectedIndex;
                if (SelectedIndex >= 0 && SelectedIndex < ListMemory.Items.Count)
                    ListMemory.Items[SelectedIndex] = new Value();
            }
        }

        private void DuplicateButton_Click(object sender, EventArgs e)
        {
            if (m_SelectedListBox == SELECTED_LIST_BOX.STACK)
            {
                int SelectedIndex = ListStack.SelectedIndex;
                if (SelectedIndex >= 0 && SelectedIndex < ListStack.Items.Count)
                {
                    object Obj = ListStack.Items[SelectedIndex];
                    if (Obj != null && Obj is Value)
                        ListStack.Items.Add((Obj as Value).Clone());
                }
            }
            else if (m_SelectedListBox == SELECTED_LIST_BOX.MEMORY)
            {
                int SelectedIndex = ListMemory.SelectedIndex;
                if (SelectedIndex >= 0 && SelectedIndex < ListMemory.Items.Count)
                {
                    object Obj = ListMemory.Items[SelectedIndex];
                    if (Obj != null && Obj is Value)
                        ListStack.Items.Add((Obj as Value).Clone());
                }
            }
        }

        private void TextBoxX_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Return && e.Control)
            {
                SaveButton_Click(null, null);
                e.Handled = true;
            }
        }

        private void TextBoxX_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == ' ')
            {
                if (sender == TextBoxX)
                {
                    TextBoxY.Focus();
                    e.Handled = true;
                }
                else if (sender == TextBoxY)
                {
                    TextBoxZ.Focus();
                    e.Handled = true;
                }
                else if (sender == TextBoxZ)
                {
                    TextBoxW.Focus();
                    e.Handled = true;
                }
            }
            else if (e.KeyChar == 13) // Enter
            {
                NewButton_Click(null, null);
                e.Handled = true;
            }
        }

        private void ListStack_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                if (ListStack.SelectedIndex >= 0 && ListStack.SelectedIndex < ListStack.Items.Count)
                    ListStack.Items.RemoveAt(ListStack.SelectedIndex);
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.Up)
            {
                if (e.Control)
                {
                    if (ListStack.SelectedIndex > 0 && ListStack.SelectedIndex < ListStack.Items.Count)
                    {
                        int Index = ListStack.SelectedIndex;
                        object Obj = ListStack.Items[Index];
                        ListStack.Items.RemoveAt(Index);
                        ListStack.Items.Insert(Index - 1, Obj);
                        ListStack.SelectedIndex = Index - 1;
                    }
                    e.Handled = true;
                }
            }
            else if (e.KeyCode == Keys.Down)
            {
                if (e.Control)
                {
                    if (ListStack.SelectedIndex >= 0 && ListStack.SelectedIndex < ListStack.Items.Count - 1)
                    {
                        int Index = ListStack.SelectedIndex;
                        object Obj = ListStack.Items[Index];
                        ListStack.Items.RemoveAt(Index);
                        ListStack.Items.Insert(Index + 1, Obj);
                        ListStack.SelectedIndex = Index + 1;
                    }
                    e.Handled = true;
                }
            }
        }

        private void ListMemory_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                if (ListMemory.SelectedIndex >= 0 && ListMemory.SelectedIndex < ListMemory.Items.Count)
                    ListMemory.Items[ListMemory.SelectedIndex] = new Value();
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.Up)
            {
                if (e.Control)
                {
                    if (ListMemory.SelectedIndex > 0 && ListMemory.SelectedIndex < ListMemory.Items.Count)
                    {
                        int Index = ListMemory.SelectedIndex;
                        object Obj = ListMemory.Items[Index];
                        ListMemory.Items.RemoveAt(Index);
                        ListMemory.Items.Insert(Index - 1, Obj);
                        ListMemory.SelectedIndex = Index - 1;
                    }
                    e.Handled = true;
                }
            }
            else if (e.KeyCode == Keys.Down)
            {
                if (e.Control)
                {
                    if (ListMemory.SelectedIndex >= 0 && ListMemory.SelectedIndex < ListMemory.Items.Count - 1)
                    {
                        int Index = ListMemory.SelectedIndex;
                        object Obj = ListMemory.Items[Index];
                        ListMemory.Items.RemoveAt(Index);
                        ListMemory.Items.Insert(Index + 1, Obj);
                        ListMemory.SelectedIndex = Index + 1;
                    }
                    e.Handled = true;
                }
            }
        }

        private void Form1_KeyDown(object sender, KeyEventArgs e)
        {
            if ((int)e.KeyCode >= (int)Keys.D0 && (int)e.KeyCode <= (int)Keys.D9)
            {
                if (e.Alt)
                {
                    int Index = (int)e.KeyCode - (int)Keys.D0;
                    if (Index == 0) Index = 9; else Index--;
                    object Obj = ListMemory.Items[Index];
                    if (Obj != null && Obj is Value && (Obj as Value).ComponentCount > 0)
                        ListStack.Items.Add((Obj as Value).Clone());
                    e.Handled = true;
                }
                else if (e.Control)
                {
                    int Index = (int)e.KeyCode - (int)Keys.D0;
                    if (Index == 0) Index = 9; else Index--;
                    if (ListStack.SelectedIndex >= 0 && ListStack.SelectedIndex < ListStack.Items.Count)
                    {
                        object Obj = ListStack.Items[ListStack.SelectedIndex];
                        if (Obj != null && Obj is Value)
                            ListMemory.Items[Index] = (Obj as Value).Clone();
                    }
                    e.Handled = true;
                }
            }
            else if ((int)e.KeyCode == 189) // '-'
            {
                if (e.Alt)
                {
                    SeparatorButton_Click(null, null);
                    e.Handled = true;
                }
            }
            else if ((int)e.KeyCode == 187) // '='
            {
                if (e.Alt)
                {
                    DuplicateButton_Click(null, null);
                    e.Handled = true;
                }
            }
            else if ((int)e.KeyCode == 192) // '~'
            {
                if (e.Alt)
                {
                    SelectColorButton_Click(null, null);
                    e.Handled = true;
                }
            }
            else if ((int)e.KeyCode == 107) // NUM+
            {
                ButtonOpClick(ButtonAdd, null);
                e.Handled = true;
            }
            else if ((int)e.KeyCode == 109) // NUM-
            {
                ButtonOpClick(ButtonSub, null);
                e.Handled = true;
            }
            else if ((int)e.KeyCode == 106) // NUM*
            {
                ButtonOpClick(ButtonMul, null);
                e.Handled = true;
            }
            else if ((int)e.KeyCode == 111) // NUM/
            {
                ButtonOpClick(ButtonDiv, null);
                e.Handled = true;
            }
            
            // Testowanie kodw klawiszy
            //MessageBox.Show(((int)e.KeyCode).ToString());
        }

        private void TextBoxX_Enter(object sender, EventArgs e)
        {
            (sender as TextBox).SelectAll();
        }

        private void Form1_Shown(object sender, EventArgs e)
        {
            TextBoxX.Focus();
        }
    }

    public class MyDragData
    {
        public Form1.SELECTED_LIST_BOX Sender;
        public Value Value;

        public MyDragData(Form1.SELECTED_LIST_BOX Sender, Value Value)
        {
            this.Sender = Sender;
            this.Value = Value;
        }
    }

    public class Value : ICloneable
    {
        private double m_x, m_y, m_z, m_w;
        private int m_ComponentCount; // 0..4

        public string Comment;

        private string m_S = null;
        private Color m_Color;

        public double x { get { return m_x; } }
        public double y { get { return m_y; } }
        public double z { get { return m_z; } }
        public double w { get { return m_w; } }
        public int ComponentCount { get { return m_ComponentCount; } }

        public double this[int i]
        {
            get
            {
                switch (i)
                {
                    case 0: return m_x;
                    case 1: return m_y;
                    case 2: return m_z;
                    case 3: return m_w;
                    default: return 0.0f;
                }
            }
        }

        public override string ToString() { return m_S; }
        public Color Color { get { return m_Color; } }

        public Value()
        {
            m_ComponentCount = 0;
            m_S = string.Empty;
            m_Color = Color.Black;
        }

        public Value(double ax)
        {
            m_x = ax;
            m_ComponentCount = 1;
            CalcS();
            CalcColor();
        }

        public Value(double ax, double ay)
        {
            m_x = ax; m_y = ay;
            m_ComponentCount = 2;
            CalcS();
            CalcColor();
        }

        public Value(double ax, double ay, double az)
        {
            m_x = ax; m_y = ay; m_z = az;
            m_ComponentCount = 3;
            CalcS();
            CalcColor();
        }

        public Value(double ax, double ay, double az, double aw)
        {
            m_x = ax; m_y = ay; m_z = az; m_w = aw;
            m_ComponentCount = 4;
            CalcS();
            CalcColor();
        }

        public Value(int ComponentCount, double ax, double ay, double az, double aw)
        {
            m_x = ax; m_y = ay; m_z = az; m_w = aw;
            m_ComponentCount = ComponentCount;
            CalcS();
            CalcColor();
        }

        const int COMPONENT_LENGTH = 9;

        private void CalcS()
        {
            System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.InvariantCulture;
            StringBuilder sb = new StringBuilder(32);
            if (ComponentCount >= 1)
            {
                sb.Append(x.ToString("G5", ci));
            }
            if (ComponentCount >= 2)
            {
                sb.Append(',');
                while (sb.Length < COMPONENT_LENGTH)
                    sb.Append(' ');
                sb.Append(y.ToString("G5", ci));
            }
            if (ComponentCount >= 3)
            {
                sb.Append(',');
                while (sb.Length < COMPONENT_LENGTH * 2)
                    sb.Append(' ');
                sb.Append(z.ToString("G5", ci));
            }
            if (ComponentCount == 4)
            {
                sb.Append(',');
                while (sb.Length < COMPONENT_LENGTH * 3)
                    sb.Append(' ');
                sb.Append(w.ToString("G5", ci));
            }
            m_S = sb.ToString();
        }

        private void CalcColor()
        {
            double fx = ComponentCount >= 1 ? x : 0; if (fx < 0.0) fx = 0.0; if (fx > 255.0) fx = 255.0;
            double fy = ComponentCount >= 2 ? y : 0; if (fy < 0.0) fy = 0.0; if (fy > 255.0) fy = 255.0;
            double fz = ComponentCount >= 3 ? z : 0; if (fz < 0.0) fz = 0.0; if (fz > 255.0) fz = 255.0;
            int R = (int)(fx * 255.0);
            int G = (int)(fy * 255.0);
            int B = (int)(fz * 255.0);
            if (R < 0) R = 0; else if (R > 255) R = 255;
            if (G < 0) G = 0; else if (G > 255) G = 255;
            if (B < 0) B = 0; else if (B > 255) B = 255;
            m_Color = Color.FromArgb(R, G, B);
        }

        public string GetShortString()
        {
            System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.InvariantCulture;
            StringBuilder sb = new StringBuilder(32);
            if (ComponentCount >= 1)
            {
                sb.Append(x.ToString("G5", ci));
            }
            if (ComponentCount >= 2)
            {
                sb.Append(", ");
                sb.Append(y.ToString("G5", ci));
            }
            if (ComponentCount >= 3)
            {
                sb.Append(", ");
                sb.Append(z.ToString("G5", ci));
            }
            if (ComponentCount == 4)
            {
                sb.Append(", ");
                sb.Append(w.ToString("G5", ci));
            }
            return sb.ToString();
        }

        public object Clone()
        {
            Value R = new Value(ComponentCount, x, y, z, w);
            R.Comment = Comment;
            return R;
        }
    }
}