using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace GameDevCalc
{
    public partial class FormFormats : Form
    {
        public FormFormats()
        {
            InitializeComponent();

            foreach (VectorFormat F in VectorFormatCollection.Formats)
            {
                FormatsListBox.Items.Add(F.GetExampleString());
            }
        }

        private void CloseButton_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.OK;
        }

        private void FormatsListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            int SelectedIndex = FormatsListBox.SelectedIndex;
            if (SelectedIndex >= 0 &&
                SelectedIndex < FormatsListBox.Items.Count &&
                SelectedIndex < VectorFormatCollection.Formats.Count)
            {
                VectorFormat F = VectorFormatCollection.Formats[SelectedIndex];
                Vector1TextBox.Text = F.Format1;
                Vector2TextBox.Text = F.Format2;
                Vector3TextBox.Text = F.Format3;
                Vector4TextBox.Text = F.Format4;
            }
        }

        private void DeleteButton_Click(object sender, EventArgs e)
        {
            int SelectedIndex = FormatsListBox.SelectedIndex;
            if (SelectedIndex >= 0 &&
                SelectedIndex < FormatsListBox.Items.Count &&
                SelectedIndex < VectorFormatCollection.Formats.Count)
            {
                if (MessageBox.Show(
                    this,
                    "Do you really want to delete this format ?",
                    "Delete format",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation) == DialogResult.Yes)
                {
                    VectorFormatCollection.Formats.RemoveAt(SelectedIndex);
                    FormatsListBox.Items.RemoveAt(SelectedIndex);
                }
            }
        }

        private void AddButton_Click(object sender, EventArgs e)
        {
            VectorFormat F = new VectorFormat();
            F.Format1 = Vector1TextBox.Text;
            F.Format2 = Vector2TextBox.Text;
            F.Format3 = Vector3TextBox.Text;
            F.Format4 = Vector4TextBox.Text;
            if (F.Validate())
            {
                VectorFormatCollection.Formats.Add(F);
                FormatsListBox.Items.Add(F.GetExampleString());
            }
            else
                MessageBox.Show(this, "Invalid format.", "Add format", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private void ModifyButton_Click(object sender, EventArgs e)
        {
            int SelectedIndex = FormatsListBox.SelectedIndex;
            if (SelectedIndex >= 0 &&
                SelectedIndex < FormatsListBox.Items.Count &&
                SelectedIndex < VectorFormatCollection.Formats.Count)
            {
                VectorFormat F = new VectorFormat();
                F.Format1 = Vector1TextBox.Text;
                F.Format2 = Vector2TextBox.Text;
                F.Format3 = Vector3TextBox.Text;
                F.Format4 = Vector4TextBox.Text;

                if (F.Validate())
                {
                    VectorFormatCollection.Formats[SelectedIndex] = F;
                    FormatsListBox.Items[SelectedIndex] = F.GetExampleString();
                }
                else
                    MessageBox.Show(this, "Invalid format.", "Modify format", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void HelpButton__Click(object sender, EventArgs e)
        {
            Form1.OnHelp();
        }
    }

    public class VectorFormat
    {
        public string Format1, Format2, Format3, Format4;

        public string GetExampleString()
        {
            return ValueToString(new Value(0.0, 0.5, 0.752, 1.0));
        }

        public string ValueToString(Value v)
        {
            if (v.ComponentCount == 0)
                return string.Empty;
            else if (v.ComponentCount == 1)
                return ValueToString_My(Format1, v);
            else if (v.ComponentCount == 2)
                return ValueToString_My(Format2, v);
            else if (v.ComponentCount == 3)
                return ValueToString_My(Format3, v);
            else // if (v.ComponentCount == 4)
                return ValueToString_My(Format4, v);
        }

        private string ValueToString_My(string Format, Value v)
        {
            StringBuilder sb = new StringBuilder();
            int Index = 0;
            for (; ; )
            {
                // Znajd '%'
                int PercentPos = Format.IndexOf('%', Index);
                // '%' nie znaleziono
                if (PercentPos == -1)
                {
                    // Przepisz do koca
                    sb.Append(Format.Substring(Index));
                    // Koniec ptli
                    break;
                }
                // Przepisz do miejsca znalezienia
                sb.Append(Format.Substring(Index, PercentPos - Index));
                // Sparsuj
                string FormatNumber;
                char FormatChar;
                int ComponentIndex;
                int Index2 = PercentPos;
                string FormattedNumber;
                if (ParsePercent(Format, ref Index2, out FormatNumber, out FormatChar, out ComponentIndex) &&
                    ComponentIndex < v.ComponentCount &&
                    GetFormattedNumber(out FormattedNumber, v[ComponentIndex], FormatNumber, FormatChar))
                {
                    sb.Append(FormattedNumber);
                    Index = Index2;
                }
                else
                {
                    // Przepisz ten jeden znak '%' i przesu do przodu o 1.
                    sb.Append(Format[PercentPos]);
                    Index++;
                }
            }
            return sb.ToString();
        }

        // Prbuje sparsowa fragment acucha z '%'.
        // Index musi na pocztku wskazywa na znak '%'.
        // Bd: zwraca false.
        private bool ParsePercent(string Format, ref int Index, out string FormatNumber, out char FormatChar, out int ComponentIndex)
        {
            // Compiler's Food
            FormatNumber = "";
            FormatChar = 'X';
            ComponentIndex = 0;

            // Spisz liczb po % (FormatNumber)
            int Index2 = Index + 1;
            while (Index < Format.Length && char.IsDigit(Format[Index2]))
                Index2++;
            // - Indeks wyszed poza zakres
            if (Index2 >= Format.Length) return false;
            FormatNumber = Format.Substring(Index + 1, Index2 - Index - 1);
            // FormatChar
            if (Index2 >= Format.Length) return false;
            FormatChar = Format[Index2];
            Index2++;
            // ComponentIndex
            if (Index2 >= Format.Length) return false;
            char ComponentChar = Format[Index2];
            Index = Index2 + 1;
            if (!ComponentCharToIndex(out ComponentIndex, ComponentChar))
                return false;

            return true;
        }

        private bool ComponentCharToIndex(out int ComponentIndex, char ComponentChar)
        {
            ComponentIndex = 0; // Compiler's Food
            switch (ComponentChar)
            {
                case 'x': case 'r': ComponentIndex = 0; return true;
                case 'y': case 'g': ComponentIndex = 1; return true;
                case 'z': case 'b': ComponentIndex = 2; return true;
                case 'w': case 'a': ComponentIndex = 3; return true;
            }
            return false;
        }

        private bool GetFormattedNumber(out string Out, double x, string FormatNumber, char FormatChar)
        {
            Out = ""; // Compiler's Food
            if (FormatNumber != string.Empty)
            {
                double Factor;
                if (!double.TryParse(FormatNumber, out Factor))
                    return false;
                x *= Factor;
            }

            switch (FormatChar)
            {
                case 'f':
                    Out = x.ToString(Form1.m_InvariantCulture);
                    // Nie ma kropki (wysza liczba cakowita) - dodaj ".0"
                    if (Out.IndexOfAny(new char[] { '.', 'e', 'E' }) == -1)
                        Out = Out + ".0";
                    break;
                case 'i':
                    Out = ((int)Math.Round(x)).ToString(Form1.m_InvariantCulture);
                    break;
                case 'x':
                    Out = Convert.ToString((int)Math.Round(x), 16);
                    Out = Out.ToUpper();
                    if (Out.Length < 2)
                        Out = "0" + Out;
                    break;
                default:
                    return false;
            }

            return true;
        }

        // W przypadku bdu konwersji zwraca NULL.
        public Value StringToValue(string s)
        {
            Value v = StringToValue_My(s, Format4, 4);
            if (v == null)
            {
                v = StringToValue_My(s, Format3, 3);
                if (v == null)
                {
                    v = StringToValue_My(s, Format2, 2);
                    if (v == null)
                        v = StringToValue_My(s, Format1, 1);
                }
            }
            return v;
        }

        // W przypadku bdu konwersji zwraca NULL.
        private Value StringToValue_My(string s, string Format, int ComponentCount)
        {
            int s_index = 0, Format_index = 0;
            double[] Data = new double[4];
            bool[] Used = new bool[4] { false, false, false, false };
            for (; ; )
            {
                // Pomi biae znaki
                while (s_index < s.Length && char.IsWhiteSpace(s[s_index]))
                    s_index++;
                while (Format_index < Format.Length && char.IsWhiteSpace(Format[Format_index]))
                    Format_index++;
                // Obydwa si skoczyy - OK
                if (s_index == s.Length && Format_index == Format.Length)
                    break;
                // Jeden si skoczy a drugi nie - bd
                if (s_index == s.Length || Format_index == Format.Length)
                    return null;
                // Procent
                if (Format[Format_index] == '%')
                {
                    int Format_index2 = Format_index;
                    string FormatNumber; char FormatChar; int ComponentIndex;
                    if (ParsePercent(Format, ref Format_index2, out FormatNumber, out FormatChar, out ComponentIndex) &&
                        ComponentIndex < ComponentCount)
                    {
                        int s_index2 = s_index;
                        if (ParseNumber(out Data[ComponentIndex], s, ref s_index2, FormatNumber, FormatChar))
                        {
                            Used[ComponentIndex] = true;
                            s_index = s_index2;
                            Format_index = Format_index2;
                        }
                        else
                            // Nie udao si sparsowa liczby wg formatu z '%' - bd
                            return null;
                    }
                    else
                    {
                        // Nie mona sparsowa '%' - potraktuj jako zwyky znak
                        // Zgadza si - OK, znak zaliczony
                        if (s[s_index] == Format[Format_index])
                        {
                            s_index++;
                            Format_index++;
                        }
                        // Nie zgadza si - bd
                        else
                            return null;
                    }
                }
                // Zwyky znaczek
                else
                {
                    // Zgadza si - OK, znak zaliczony
                    if (s[s_index] == Format[Format_index])
                    {
                        s_index++;
                        Format_index++;
                    }
                    // Nie zgadza si - bd
                    else
                        return null;
                }
            }

            // Nie wszystkie komponenty uyte - bd
            for (int i = 0; i < ComponentCount; i++)
                if (!Used[i])
                    return null;

            return new Value(ComponentCount, Data[0], Data[1], Data[2], Data[3]);
        }

        private bool ParseNumber(out double x, string s, ref int s_index, string FormatNumber, char FormatChar)
        {
            x = 0.0f; // Compiler's Food
            int start_index = s_index;
            while (s_index < s.Length)
            {
                if (FormatChar == 'x' && s_index >= start_index + 2)
                    break;
                if (s[s_index] == '-' || s[s_index] == '+' || (s[s_index] >= '0' && s[s_index] <= '9'))
                    s_index++;
                else
                {
                    if (FormatChar == 'x')
                    {
                        if (s[s_index] >= 'a' && s[s_index] <= 'z' || s[s_index] >= 'A' && s[s_index] <= 'Z')
                           s_index++;
                        else
                            break;
                    }
                    else if (FormatChar == 'f')
                    {
                        if (s[s_index] == 'e' || s[s_index] == 'E' || s[s_index] == '.')
                            s_index++;
                        else
                            break;
                    }
                    else
                        break;
                }
            }

            if (FormatChar == 'f')
            {
                if (!double.TryParse(
                    s.Substring(start_index, s_index - start_index),
                    System.Globalization.NumberStyles.Number,
                    Form1.m_InvariantCulture,
                    out x))
                    return false;
                if (FormatNumber.Length > 0)
                {
                    double FormatNumberD;
                    if (double.TryParse(FormatNumber, System.Globalization.NumberStyles.Number, Form1.m_InvariantCulture, out FormatNumberD))
                        x /= FormatNumberD;
                    else
                        return false;
                }
            }
            else if (FormatChar == 'i')
            {
                int ix;
                if (!int.TryParse(s.Substring(start_index, s_index - start_index), out ix))
                    return false;
                x = (double)ix;
                if (FormatNumber.Length > 0)
                {
                    int FormatNumberD;
                    if (int.TryParse(FormatNumber, out FormatNumberD))
                        x /= FormatNumberD;
                    else
                        return false;
                }
            }
            else if (FormatChar == 'x')
            {
                try
                {
                    x = Convert.ToInt32(s.Substring(start_index, s_index - start_index), 16);
                    if (FormatNumber.Length > 0)
                    {
                        int FormatNumberD;
                        if (int.TryParse(FormatNumber, out FormatNumberD))
                            x /= FormatNumberD;
                        else
                            return false;
                    }
                }
                catch (Exception)
                {
                    return false;
                }
            }
            else
                return false;
            return true;
        }

        public bool Validate()
        {
            if (!Validate_My(Format1, 1)) return false;
            if (!Validate_My(Format2, 2)) return false;
            if (!Validate_My(Format3, 3)) return false;
            if (!Validate_My(Format4, 4)) return false;
            return true;
        }

        bool Validate_My(string Format, int ComponentCount)
        {
            int Index = 0, PercentPos;
            string FormatNumber;
            char FormatChar;
            int ComponentIndex;
            bool[] Used = new bool[4] { false, false, false, false };
            while (Index < Format.Length)
            {
                PercentPos = Format.IndexOf('%', Index);
                if (PercentPos == -1)
                    break;
                if (ParsePercent(Format, ref PercentPos, out FormatNumber, out FormatChar, out ComponentIndex))
                {
                    if (ComponentIndex >= ComponentCount)
                        return false;
                    if (Used[ComponentIndex])
                        return false;
                    Used[ComponentIndex] = true;
                    Index = PercentPos;
                }
                else
                    Index++;
            }

            for (int i = 0; i < ComponentCount; i++)
                if (!Used[i]) return false;
            return true;
        }
    }

    public static class VectorFormatCollection
    {
        public static List<VectorFormat> Formats = new List<VectorFormat>();
    }
}